#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date

from holidays.calendars.gregorian import JAN, MAR, MAY, JUN, OCT, NOV, DEC


class InternationalHolidays:
    """
    International holidays.
    """

    def _add_africa_day(self, name):
        """
        Add Africa Day (May 25th)

        Africa Day (formerly African Freedom Day and African Liberation Day)
        is the annual commemoration of the foundation of the Organisation
        of African Unity on 25 May 1963.
        https://en.wikipedia.org/wiki/Africa_Day
        """
        return self._add_holiday(name, MAY, 25)

    def _add_childrens_day(self, name, variation="JUN"):
        """
        Add International Children's Day (June 1).

        In 1925, International Children's Day was first proclaimed in Geneva
        during the World Conference on Child Welfare. Since 1950, it is
        celebrated on June 1 in many countries.

        As such, this entry currently defaults to June 1, though this also
        supports another internationally adopted variant, November 20th.
        https://en.wikipedia.org/wiki/Children's_Day
        """
        if variation == "JUN":
            return self._add_holiday(name, JUN, 1)
        elif variation == "NOV":
            return self._add_holiday(name, NOV, 20)
        else:
            raise ValueError(
                f"Unknown variaton name: {variation}. "
                "This entry currently supports `JUN` and `NOV` variation only."
            )

    def _add_columbus_day(self, name):
        """
        Add Columbus Day (October 12th)

        Columbus Day is a national holiday which officially celebrates the
        anniversary of Christopher Columbus's arrival in the Americas.
        https://en.wikipedia.org/wiki/Columbus_Day
        """
        return self._add_holiday(name, OCT, 12)

    def _add_labor_day(self, name):
        """
        Add International Workers' Day (May 1st)

        International Workers' Day, also known as Labour Day, is a celebration
        of labourers and the working classes that is promoted by the
        international labour movement.
        https://en.wikipedia.org/wiki/International_Workers%27_Day
        """
        return self._add_holiday(name, MAY, 1)

    def _add_labor_day_two(self, name):
        """
        Add International Workers' Day Two (May 2nd)

        https://en.wikipedia.org/wiki/International_Workers%27_Day
        """
        return self._add_holiday(name, MAY, 2)

    def _add_labor_day_three(self, name):
        """
        Add International Workers' Day Three (May 3rd)

        https://en.wikipedia.org/wiki/International_Workers%27_Day
        """
        return self._add_holiday(name, MAY, 3)

    def _add_new_years_day(self, name) -> date:
        """
        Add New Year's Day (January 1st).

        New Year's Day is a festival observed in most of the world on
        1 January, the first day of the year in the modern Gregorian calendar.
        https://en.wikipedia.org/wiki/New_Year%27s_Day
        """
        return self._add_holiday(name, JAN, 1)

    def _add_new_years_day_two(self, name) -> date:
        """
        Add New Year's Day Two (January 2nd).

        New Year's Day is a festival observed in most of the world on
        1 January, the first day of the year in the modern Gregorian calendar.
        https://en.wikipedia.org/wiki/New_Year%27s_Day
        """
        return self._add_holiday(name, JAN, 2)

    def _add_new_years_day_three(self, name) -> date:
        """
        Add New Year's Day Three (January 3rd).

        New Year's Day is a festival observed in most of the world on
        1 January, the first day of the year in the modern Gregorian calendar.
        https://en.wikipedia.org/wiki/New_Year%27s_Day
        """
        return self._add_holiday(name, JAN, 3)

    def _add_new_years_day_four(self, name) -> date:
        """
        Add New Year's Day Four (January 4th).

        New Year's Day is a festival observed in most of the world on
        1 January, the first day of the year in the modern Gregorian calendar.
        https://en.wikipedia.org/wiki/New_Year%27s_Day
        """
        return self._add_holiday(name, JAN, 4)

    def _add_remembrance_day(self, name):
        """
        Add Remembrance Day / Armistice Day (Nov 11th)

        It's a memorial day since the end of the First World War in 1919
        to honour armed forces members who have died in the line of duty.
        https://en.wikipedia.org/wiki/Remembrance_Day
        """
        return self._add_holiday(name, NOV, 11)

    def _add_new_years_eve(self, name) -> date:
        """
        Add New Year's Eve (December 31st).

        In the Gregorian calendar, New Year's Eve, also known as Old Year's
        Day or Saint Sylvester's Day in many countries, is the evening or the
        entire day of the last day of the year, on 31 December.
        https://en.wikipedia.org/wiki/New_Year%27s_Eve
        """
        return self._add_holiday(name, DEC, 31)

    def _add_womens_day(self, name):
        """
        Add International Women's Day (March 8th).

        International Women's Day is a global holiday celebrated as a focal
        point in the women's rights movement, bringing attention to issues
        such as gender equality, reproductive rights, and violence and abuse
        against women.
        https://en.wikipedia.org/wiki/International_Women%27s_Day
        """
        return self._add_holiday(name, MAR, 8)

    def _add_world_war_two_victory_day(self, name):
        """
        Add Victory Day (May 9th)

        Victory Day is a holiday that commemorates the victory over Nazi
        Germany in 1945.
        https://en.wikipedia.org/wiki/Victory_Day_(9_May)
        """
        return self._add_holiday(name, MAY, 9)



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

# flake8: noqa: F401

from datetime import date
from datetime import timedelta as td

from holidays.calendars.buddhist import _CustomBuddhistCalendar, _BuddhistLunisolar
from holidays.calendars.chinese import _CustomChineseCalendar, _ChineseLunisolar
from holidays.calendars.custom import _CustomCalendar
from holidays.calendars.gregorian import GREGORIAN_CALENDAR
from holidays.calendars.hebrew import _HebrewLunisolar
from holidays.calendars.hindu import _HinduLunisolar, _CustomHinduCalendar
from holidays.calendars.islamic import _CustomIslamicCalendar, _IslamicLunar
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.calendars.thai import _ThaiLunisolar, KHMER_CALENDAR, THAI_CALENDAR



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, APR, JUN, JUL, OCT
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, IslamicHolidays, InternationalHolidays


class Egypt(HolidayBase, ChristianHolidays, IslamicHolidays, InternationalHolidays):
    # Holidays here are estimates, it is common for the day to be pushed
    # if falls in a weekend, although not a rule that can be implemented.
    # Holidays after 2020: the following four moving date holidays whose exact
    # date is announced yearly are estimated (and so denoted):
    # - Eid El Fetr*
    # - Eid El Adha*
    # - Arafat Day*
    # - Moulad El Naby*
    # *only if hijri-converter library is installed, otherwise a warning is
    #  raised that this holiday is missing. hijri-converter requires
    #  Python >= 3.6
    # is_weekend function is there, however not activated for accuracy.

    country = "EG"
    default_language = "ar"
    # Estimated label.
    estimated_label = tr("(تقدير*) *%s")
    supported_languages = ("ar", "en_US")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)

        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day
        self._add_new_years_day(tr("رأس السنة الميلادية"))

        # Coptic Christmas
        self._add_christmas_day(tr("عيد الميلاد المجيد (تقويم قبطي)"))

        if year >= 2012:
            # January 25th Revolution
            self._add_holiday(tr("عيد ثورة 25 يناير"), JAN, 25)
        elif year >= 2009:
            # National Police Day
            self._add_holiday(tr("عيد الشرطة"), JAN, 25)

        # Coptic Easter - Orthodox Easter
        self._add_easter_sunday(tr("عيد الفصح القبطي"))
        self._add_easter_monday(tr("شم النسيم"))  # Spring Festival

        if year > 1982:
            # Sinai Libration Day
            self._add_holiday(tr("عيد تحرير سيناء"), APR, 25)

        # Labour Day
        self._add_labor_day(tr("عيد العمال"))

        # Armed Forces Day
        self._add_holiday(tr("عيد القوات المسلحة"), OCT, 6)

        if year >= 2014:
            # June 30 Revolution Day
            self._add_holiday(tr("عيد ثورة 30 يونيو"), JUN, 30)

        if year > 1952:
            # July 23 Revolution Day
            self._add_holiday(tr("عيد ثورة 23 يوليو"), JUL, 23)

        # Eid al-Fitr - Feast Festive
        self._add_eid_al_fitr_day(tr("عيد الفطر"))
        # Eid al-Fitr Holiday
        self._add_eid_al_fitr_day_two(tr("عطلة عيد الفطر"))
        self._add_eid_al_fitr_day_three(tr("عطلة عيد الفطر"))

        # Arafat Day
        self._add_arafah_day(tr("يوم عرفة"))

        # Eid al-Adha - Scarfice Festive
        self._add_eid_al_adha_day(tr("عيد الأضحى"))
        # Eid al-Adha Holiday
        self._add_eid_al_adha_day_two(tr("عطلة عيد الأضحى"))
        self._add_eid_al_adha_day_three(tr("عطلة عيد الأضحى"))

        # Islamic New Year
        self._add_islamic_new_year_day(tr("رأس السنة الهجرية"))

        # Prophet Muhammad's Birthday
        self._add_mawlid_day(tr("عيد المولد النبوي"))


class EG(Egypt):
    pass


class EGY(Egypt):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import MAR, APR, MAY, JUN, JUL, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class France(HolidayBase, ChristianHolidays, InternationalHolidays):
    """Official French holidays.

    Some provinces have specific holidays, only those are included in the
    PROVINCES, because these provinces have different administrative status,
    which makes it difficult to enumerate.

    For religious holidays usually happening on Sundays (Easter, Pentecost),
    only the following Monday is considered a holiday.

    Primary sources:
        https://fr.wikipedia.org/wiki/Fêtes_et_jours_fériés_en_France
        https://www.service-public.fr/particuliers/vosdroits/F2405
    """

    country = "FR"
    default_language = "fr"
    supported_languages = ("en_US", "fr", "uk")
    subdivisions = (
        "BL",  # Saint Barthelemy.
        "GES",  # Alsace, Champagne-Ardenne, Lorraine(Moselle).
        "GP",  # Guadeloupe.
        "GY",  # Guyane.
        "MF",  # Saint Martin.
        "MQ",  # Martinique.
        "NC",  # Nouvelle-Calédonie,
        "PF",  # Polynésie Française.
        "RE",  # Reunion.
        "WF",  # Wallis-et-Futuna.
        "YT",  # Mayotte.
    )

    _deprecated_subdivisions = (
        "Alsace-Moselle",
        "Guadeloupe",
        "Guyane",
        "La Réunion",
        "Martinique",
        "Mayotte",
        "Métropole",
        "Nouvelle-Calédonie",
        "Polynésie Française",
        "Saint-Barthélémy",
        "Saint-Martin",
        "Wallis-et-Futuna",
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # Civil holidays.
        if year >= 1811:
            # New Year's Day.
            self._add_new_years_day(tr("Jour de l'an"))

        if year >= 1919:
            self._add_labor_day(
                # Labor Day.
                tr("Fête du Travail")
                if year >= 1948
                # Labor and Social Concord Day.
                else tr("Fête du Travail et de la Concorde sociale")
            )

        if 1953 <= year <= 1959 or year >= 1982:
            # Victory Day.
            self._add_holiday(tr("Fête de la Victoire"), MAY, 8)

        if year >= 1880:
            # National Day.
            self._add_holiday(tr("Fête nationale"), JUL, 14)

        if year >= 1918:
            # Armistice Day.
            self._add_holiday(tr("Armistice"), NOV, 11)

        # Religious holidays.

        if year >= 1886:
            # Easter Monday.
            self._add_easter_monday(tr("Lundi de Pâques"))

            if year not in {2005, 2006, 2007}:
                # Whit Monday.
                self._add_whit_monday(tr("Lundi de Pentecôte"))

        if year >= 1802:
            # Ascension Day.
            self._add_ascension_thursday(tr("Ascension"))
            # Assumption Day.
            self._add_assumption_of_mary_day(tr("Assomption"))
            # All Saints' Day.
            self._add_all_saints_day(tr("Toussaint"))
            # Christmas Day.
            self._add_christmas_day(tr("Noël"))

        if self.subdiv == "Alsace-Moselle":
            self._add_subdiv_ges_holidays()
        elif self.subdiv == "Guadeloupe":
            self._add_subdiv_gp_holidays()
        elif self.subdiv == "Guyane":
            self._add_subdiv_gy_holidays()
        elif self.subdiv == "La Réunion":
            self._add_subdiv_re_holidays()
        elif self.subdiv == "Martinique":
            self._add_subdiv_mq_holidays()
        elif self.subdiv == "Mayotte":
            self._add_subdiv_yt_holidays()
        elif self.subdiv == "Nouvelle-Calédonie":
            self._add_subdiv_nc_holidays()
        elif self.subdiv == "Polynésie Française":
            self._add_subdiv_pf_holidays()
        elif self.subdiv == "Saint-Barthélémy":
            self._add_subdiv_bl_holidays()
        elif self.subdiv == "Saint-Martin":
            self._add_subdiv_mf_holidays()
        elif self.subdiv == "Wallis-et-Futuna":
            self._add_subdiv_wf_holidays()

    # Saint Barthelemy.
    def _add_subdiv_bl_holidays(self):
        # Abolition of slavery.
        self._add_holiday(tr("Abolition de l'esclavage"), OCT, 9)

    # Alsace, Champagne-Ardenne, Lorraine(Moselle).
    def _add_subdiv_ges_holidays(self):
        # Good Friday.
        self._add_good_friday(tr("Vendredi saint"))

        # Saint Stephen's Day.
        self._add_christmas_day_two(tr("Saint Étienne"))

    # Guadeloupe.
    def _add_subdiv_gp_holidays(self):
        # Good Friday.
        self._add_good_friday(tr("Vendredi saint"))

        # Mi-Careme.
        self._add_holiday(tr("Mi-Carême"), self._easter_sunday + td(days=-24))

        # Abolition of slavery.
        self._add_holiday(tr("Abolition de l'esclavage"), MAY, 27)

        # Feast of Victor Schoelcher.
        self._add_holiday(tr("Fête de Victor Schoelcher"), JUL, 21)

    # Guyane.
    def _add_subdiv_gy_holidays(self):
        # Abolition of slavery.
        self._add_holiday(tr("Abolition de l'esclavage"), JUN, 10)

    # Saint Martin.
    def _add_subdiv_mf_holidays(self):
        if self._year >= 2018:
            # Abolition of slavery.
            self._add_holiday(tr("Abolition de l'esclavage"), MAY, 28)

    # Martinique.
    def _add_subdiv_mq_holidays(self):
        # Good Friday.
        self._add_good_friday(tr("Vendredi saint"))

        # Abolition of slavery.
        self._add_holiday(tr("Abolition de l'esclavage"), MAY, 22)

        # Feast of Victor Schoelcher.
        self._add_holiday(tr("Fête de Victor Schoelcher"), JUL, 21)

    # New Caledonia.
    def _add_subdiv_nc_holidays(self):
        # Citizenship Day.
        self._add_holiday(tr("Fête de la Citoyenneté"), SEP, 24)

    # French Polynesia.
    def _add_subdiv_pf_holidays(self):
        # Good Friday.
        self._add_good_friday(tr("Vendredi saint"))

        # Missionary Day.
        self._add_holiday(tr("Arrivée de l'Évangile"), MAR, 5)

        # Internal Autonomy Day.
        self._add_holiday(tr("Fête de l'autonomie"), JUN, 29)

    # Reunion.
    def _add_subdiv_re_holidays(self):
        if self._year >= 1981:
            # Abolition of slavery.
            self._add_holiday(tr("Abolition de l'esclavage"), DEC, 20)

    #  Wallis and Futuna.
    def _add_subdiv_wf_holidays(self):
        # Feast of Saint Peter Chanel.
        self._add_holiday(tr("Saint Pierre Chanel"), APR, 28)

        # Festival of the territory.
        self._add_holiday(tr("Fête du Territoire"), JUL, 29)

    # Mayotte.
    def _add_subdiv_yt_holidays(self):
        # Abolition of slavery.
        self._add_holiday(tr("Abolition de l'esclavage"), APR, 27)


class FR(France):
    """FR is also used by dateutil (Friday), so be careful with this one."""

    pass


class FRA(France):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.countries.united_states import US


class HolidaysMP(US):
    # Alias of a US subdivision that is also officially assigned its own
    # country code in ISO 3166-1.  See
    # https://en.wikipedia.org/wiki/ISO_3166-2:US#Subdivisions_included_in_ISO_3166-1

    country = "MP"
    subdivisions = ()  # Override US subdivisions.

    def _populate(self, year: int) -> None:
        self.subdiv = "MP"
        super()._populate(year)


class MP(HolidaysMP):
    pass


class MNP(HolidaysMP):
    pass


class NorthernMarianaIslands(HolidaysMP):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Provinces completed by Henrik Sozzi <henrik_sozzi@hotmail.com>
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    MAY,
    JUN,
    JUL,
    AUG,
    SEP,
    OCT,
    NOV,
    DEC,
    TUE,
    SUN,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Italy(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "IT"
    # Reference: https://it.wikipedia.org/wiki/Province_d%27Italia
    # Please maintain in alphabetical order for easy updating in the future
    # The alphabetical order is except cities of provinces with multiple head
    # cities that directly follows the main province id like BT, Barletta,
    # Andria, Trani, for easily grouping them.
    # In that case if you use the 2 char id you'll take the first Santo
    # Patrono defined. If you want one specific you'll have to use
    # the full name of the city like "Andria" instead of "BT".
    subdivisions = (
        # Provinces.
        "AG",
        "AL",
        "AN",
        "AO",
        "AP",
        "AQ",
        "AR",
        "AT",
        "AV",
        "BA",
        "BG",
        "BI",
        "BL",
        "BN",
        "BO",
        "BR",
        "BS",
        "BT",
        "BZ",
        "CA",
        "CB",
        "CE",
        "CH",
        "CL",
        "CN",
        "CO",
        "CR",
        "CS",
        "CT",
        "CZ",
        "EN",
        "FC",
        "FE",
        "FG",
        "FI",
        "FM",
        "FR",
        "GE",
        "GO",
        "GR",
        "IM",
        "IS",
        "KR",
        "LC",
        "LE",
        "LI",
        "LO",
        "LT",
        "LU",
        "MB",
        "MC",
        "ME",
        "MI",
        "MN",
        "MO",
        "MS",
        "MT",
        "NA",
        "NO",
        "NU",
        "OR",
        "PA",
        "PC",
        "PD",
        "PE",
        "PG",
        "PI",
        "PN",
        "PO",
        "PR",
        "PT",
        "PU",
        "PV",
        "PZ",
        "RA",
        "RC",
        "RE",
        "RG",
        "RI",
        "RM",
        "RN",
        "RO",
        "SA",
        "SI",
        "SO",
        "SP",
        "SR",
        "SS",
        "SU",
        "SV",
        "TA",
        "TE",
        "TN",
        "TO",
        "TP",
        "TR",
        "TS",
        "TV",
        "UD",
        "VA",
        "VB",
        "VC",
        "VE",
        "VI",
        "VR",
        "VT",
        "VV",
        # Cities.
        "Andria",
        "Barletta",
        "Cesena",
        "Forli",
        "Pesaro",
        "Trani",
        "Urbino",
    )

    _deprecated_subdivisions = ("Forlì",)

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("Capodanno")

        # Epiphany.
        self._add_epiphany_day("Epifania del Signore")

        # Easter Sunday.
        self._add_easter_sunday("Pasqua di Resurrezione")

        # Easter Monday.
        self._add_easter_monday("Lunedì dell'Angelo")

        if year >= 1946:
            # Liberation Day.
            self._add_holiday("Festa della Liberazione", APR, 25)

        # Labor Day.
        self._add_labor_day("Festa dei Lavoratori")

        if year >= 1948:
            # Republic Day.
            self._add_holiday("Festa della Repubblica", JUN, 2)

        # Assumption Of Mary Day.
        self._add_assumption_of_mary_day("Assunzione della Vergine")

        # All Saints' Day.
        self._add_all_saints_day("Tutti i Santi")

        # Immaculate Conception Day.
        self._add_immaculate_conception_day("Immacolata Concezione")

        # Christmas Day.
        self._add_christmas_day("Natale")

        self._add_christmas_day_two("Santo Stefano")

        if self.subdiv == "Forlì":
            self._add_subdiv_forli_holidays()

    # Provinces holidays.
    # https://it.wikipedia.org/wiki/Santi_patroni_cattolici_delle_citt%C3%A0_capoluogo_di_provincia_italiane
    # Please maintain in alphabetical order for easy updating in the future.

    def _add_subdiv_ag_holidays(self):
        self._add_holiday("San Gerlando", FEB, 25)

    def _add_subdiv_al_holidays(self):
        self._add_holiday("San Baudolino", NOV, 10)

    def _add_subdiv_an_holidays(self):
        self._add_holiday("San Ciriaco", MAY, 4)

    def _add_subdiv_ao_holidays(self):
        self._add_holiday("San Grato", SEP, 7)

    def _add_subdiv_ap_holidays(self):
        self._add_holiday("Sant'Emidio", AUG, 5)

    def _add_subdiv_aq_holidays(self):
        self._add_holiday("San Massimo D'Aveia", JUN, 10)

    def _add_subdiv_ar_holidays(self):
        self._add_holiday("San Donato D'Arezzo", AUG, 7)

    def _add_subdiv_at_holidays(self):
        self._add_holiday("San Secondo di Asti", self._get_nth_weekday_of_month(1, TUE, MAY))

    def _add_subdiv_av_holidays(self):
        self._add_holiday("San Modestino", FEB, 14)

    def _add_subdiv_ba_holidays(self):
        self._add_holiday("San Nicola", DEC, 6)

    def _add_subdiv_bg_holidays(self):
        self._add_holiday("Sant'Alessandro di Bergamo", AUG, 26)

    def _add_subdiv_bi_holidays(self):
        self._add_christmas_day_two("Santo Stefano")

    def _add_subdiv_bl_holidays(self):
        self._add_holiday("San Martino", NOV, 11)

    def _add_subdiv_bn_holidays(self):
        self._add_holiday("San Bartolomeo apostolo", AUG, 24)

    def _add_subdiv_bo_holidays(self):
        self._add_holiday("San Petronio", OCT, 4)

    def _add_subdiv_br_holidays(self):
        self._add_holiday(
            "San Teodoro d'Amasea e San Lorenzo da Brindisi",
            self._get_nth_weekday_of_month(1, SUN, SEP),
        )

    def _add_subdiv_bs_holidays(self):
        self._add_holiday("Santi Faustino e Giovita", FEB, 15)

    def _add_subdiv_bt_holidays(self):
        self._add_holiday("San Nicola Pellegrino", MAY, 3)
        self._add_holiday("San Riccardo di Andria", self._get_nth_weekday_of_month(3, SUN, SEP))
        self._add_holiday("San Ruggero", DEC, 30)

    def _add_subdiv_bz_holidays(self):
        self._add_whit_monday("Lunedì di Pentecoste")
        self._add_assumption_of_mary_day("Maria Santissima Assunta")

    def _add_subdiv_ca_holidays(self):
        self._add_holiday("San Saturnino di Cagliari", OCT, 30)

    def _add_subdiv_cb_holidays(self):
        self._add_saint_georges_day("San Giorgio")

    def _add_subdiv_ce_holidays(self):
        self._add_holiday("San Sebastiano", JAN, 20)

    def _add_subdiv_ch_holidays(self):
        self._add_holiday("San Giustino di Chieti", MAY, 11)

    def _add_subdiv_cl_holidays(self):
        self._add_holiday("San Michele Arcangelo", SEP, 29)

    def _add_subdiv_cn_holidays(self):
        self._add_holiday("San Michele Arcangelo", SEP, 29)

    def _add_subdiv_co_holidays(self):
        self._add_holiday("Sant'Abbondio", AUG, 31)

    def _add_subdiv_cr_holidays(self):
        self._add_holiday("Sant'Omobono", NOV, 13)

    def _add_subdiv_cs_holidays(self):
        self._add_holiday("Madonna del Pilerio", FEB, 12)

    def _add_subdiv_ct_holidays(self):
        self._add_holiday("Sant'Agata", FEB, 5)

    def _add_subdiv_cz_holidays(self):
        self._add_holiday("San Vitaliano", JUL, 16)

    def _add_subdiv_en_holidays(self):
        self._add_holiday("Madonna della Visitazione", JUL, 2)

    def _add_subdiv_fc_holidays(self):
        self._add_holiday("Madonna del Fuoco", FEB, 4)
        self._add_saint_johns_day("San Giovanni Battista")

    def _add_subdiv_fe_holidays(self):
        self._add_saint_georges_day("San Giorgio")

    def _add_subdiv_fg_holidays(self):
        self._add_holiday("Madonna dei Sette Veli", MAR, 22)

    def _add_subdiv_fi_holidays(self):
        self._add_saint_johns_day("San Giovanni Battista")

    def _add_subdiv_fm_holidays(self):
        aug_15 = self._add_assumption_of_mary_day("Maria Santissima Assunta")
        self._add_holiday("Maria Santissima Assunta", aug_15 + td(days=+1))

    def _add_subdiv_fr_holidays(self):
        self._add_holiday("San Silverio", JUN, 20)

    def _add_subdiv_ge_holidays(self):
        self._add_saint_johns_day("San Giovanni Battista")

    def _add_subdiv_go_holidays(self):
        self._add_holiday("Santi Ilario e Taziano", MAR, 16)

    def _add_subdiv_gr_holidays(self):
        self._add_holiday("San Lorenzo", AUG, 10)

    def _add_subdiv_im_holidays(self):
        self._add_holiday("San Leonardo da Porto Maurizio", NOV, 26)

    def _add_subdiv_is_holidays(self):
        self._add_holiday("San Pietro Celestino", MAY, 19)

    def _add_subdiv_kr_holidays(self):
        self._add_holiday("San Dionigi", OCT, 9)

    def _add_subdiv_lc_holidays(self):
        self._add_holiday("San Nicola", DEC, 6)

    def _add_subdiv_le_holidays(self):
        self._add_holiday("Sant'Oronzo", AUG, 26)

    def _add_subdiv_li_holidays(self):
        self._add_holiday("Santa Giulia", MAY, 22)

    def _add_subdiv_lo_holidays(self):
        self._add_holiday("San Bassiano", JAN, 19)

    def _add_subdiv_lt_holidays(self):
        self._add_holiday("San Marco evangelista", APR, 25)

    def _add_subdiv_lu_holidays(self):
        self._add_holiday("San Paolino di Lucca", JUL, 12)

    def _add_subdiv_mb_holidays(self):
        self._add_saint_johns_day("San Giovanni Battista")

    def _add_subdiv_mc_holidays(self):
        self._add_holiday("San Giuliano l'ospitaliere", AUG, 31)

    def _add_subdiv_me_holidays(self):
        self._add_holiday("Madonna della Lettera", JUN, 3)

    def _add_subdiv_mi_holidays(self):
        self._add_holiday("Sant'Ambrogio", DEC, 7)

    def _add_subdiv_mn_holidays(self):
        self._add_holiday("Sant'Anselmo da Baggio", MAR, 18)

    def _add_subdiv_mo_holidays(self):
        self._add_holiday("San Geminiano", JAN, 31)

    def _add_subdiv_ms_holidays(self):
        self._add_holiday("San Francesco d'Assisi", OCT, 4)

    def _add_subdiv_mt_holidays(self):
        self._add_holiday("Madonna della Bruna", JUL, 2)

    def _add_subdiv_na_holidays(self):
        self._add_holiday("San Gennaro", SEP, 19)

    def _add_subdiv_no_holidays(self):
        self._add_holiday("San Gaudenzio", JAN, 22)

    def _add_subdiv_nu_holidays(self):
        self._add_holiday("Nostra Signora della Neve", AUG, 5)

    def _add_subdiv_or_holidays(self):
        self._add_holiday("Sant'Archelao", FEB, 13)

    def _add_subdiv_pa_holidays(self):
        self._add_holiday("San Giovanni", JUL, 15)

    def _add_subdiv_pc_holidays(self):
        self._add_holiday("Sant'Antonino di Piacenza", JUL, 4)

    def _add_subdiv_pd_holidays(self):
        self._add_holiday("Sant'Antonio di Padova", JUN, 13)

    def _add_subdiv_pe_holidays(self):
        self._add_holiday("San Cetteo", OCT, 10)

    def _add_subdiv_pg_holidays(self):
        self._add_holiday("Sant'Ercolano e San Lorenzo", JAN, 29)

    def _add_subdiv_pi_holidays(self):
        self._add_holiday("San Ranieri", JUN, 17)

    def _add_subdiv_pn_holidays(self):
        self._add_holiday("San Marco Evangelista", APR, 25)
        self._add_nativity_of_mary_day("Madonna delle Grazie")

    def _add_subdiv_po_holidays(self):
        self._add_christmas_day_two("Santo Stefano")

    def _add_subdiv_pr_holidays(self):
        self._add_holiday("Sant'Ilario di Poitiers", JAN, 13)

    def _add_subdiv_pt_holidays(self):
        self._add_saint_james_day("San Jacopo")

    def _add_subdiv_pu_holidays(self):
        self._add_holiday("San Crescentino", JUN, 1)
        self._add_holiday("San Terenzio di Pesaro", SEP, 24)

    def _add_subdiv_pv_holidays(self):
        self._add_holiday("San Siro", DEC, 9)

    def _add_subdiv_pz_holidays(self):
        self._add_holiday("San Gerardo di Potenza", MAY, 30)

    def _add_subdiv_ra_holidays(self):
        self._add_holiday("Sant'Apollinare", JUL, 23)

    def _add_subdiv_rc_holidays(self):
        self._add_saint_georges_day("San Giorgio")

    def _add_subdiv_re_holidays(self):
        self._add_holiday("San Prospero Vescovo", NOV, 24)

    def _add_subdiv_rg_holidays(self):
        self._add_saint_georges_day("San Giorgio")

    def _add_subdiv_ri_holidays(self):
        self._add_holiday("Santa Barbara", DEC, 4)

    def _add_subdiv_rm_holidays(self):
        self._add_saints_peter_and_paul_day("Santi Pietro e Paolo")

    def _add_subdiv_rn_holidays(self):
        self._add_holiday("San Gaudenzio", OCT, 14)

    def _add_subdiv_ro_holidays(self):
        self._add_holiday("San Bellino", NOV, 26)

    def _add_subdiv_sa_holidays(self):
        self._add_holiday("San Matteo Evangelista", SEP, 21)

    def _add_subdiv_si_holidays(self):
        self._add_holiday("Sant'Ansano", DEC, 1)

    def _add_subdiv_so_holidays(self):
        self._add_holiday("San Gervasio e San Protasio", JUN, 19)

    def _add_subdiv_sp_holidays(self):
        self._add_saint_josephs_day("San Giuseppe")

    def _add_subdiv_sr_holidays(self):
        self._add_holiday("Santa Lucia", DEC, 13)

    def _add_subdiv_ss_holidays(self):
        self._add_holiday("San Nicola", DEC, 6)

    def _add_subdiv_su_holidays(self):
        self._add_holiday(
            "San Ponziano", self._get_nth_weekday_of_month(2, SUN, MAY) + td(days=+4)
        )

    def _add_subdiv_sv_holidays(self):
        self._add_holiday("Nostra Signora della Misericordia", MAR, 18)

    def _add_subdiv_ta_holidays(self):
        self._add_holiday("San Cataldo", MAY, 10)

    def _add_subdiv_te_holidays(self):
        self._add_holiday("San Berardo da Pagliara", DEC, 19)

    def _add_subdiv_tn_holidays(self):
        self._add_holiday("San Vigilio", JUN, 26)

    def _add_subdiv_to_holidays(self):
        self._add_saint_johns_day("San Giovanni Battista")

    def _add_subdiv_tp_holidays(self):
        self._add_holiday("Sant'Alberto degli Abati", AUG, 7)

    def _add_subdiv_tr_holidays(self):
        self._add_holiday("San Valentino", FEB, 14)

    def _add_subdiv_ts_holidays(self):
        self._add_holiday("San Giusto", NOV, 3)

    def _add_subdiv_tv_holidays(self):
        self._add_holiday("San Liberale", APR, 27)

    def _add_subdiv_ud_holidays(self):
        self._add_holiday("Santi Ermacora e Fortunato", JUL, 12)

    def _add_subdiv_va_holidays(self):
        self._add_holiday("San Vittore il Moro", MAY, 8)

    def _add_subdiv_vb_holidays(self):
        self._add_holiday("San Vittore il Moro", MAY, 8)

    def _add_subdiv_vc_holidays(self):
        self._add_holiday("Sant'Eusebio di Vercelli", AUG, 1)

    def _add_subdiv_ve_holidays(self):
        self._add_holiday("San Marco Evangelista", APR, 25)

    def _add_subdiv_vi_holidays(self):
        self._add_holiday("San Marco", APR, 25)

    def _add_subdiv_vr_holidays(self):
        self._add_holiday("San Zeno", MAY, 21)

    def _add_subdiv_vt_holidays(self):
        self._add_holiday("Santa Rosa da Viterbo", SEP, 4)

    def _add_subdiv_vv_holidays(self):
        self._add_holiday("San Leoluca", MAR, 1)

    def _add_subdiv_andria_holidays(self):
        self._add_holiday("San Riccardo di Andria", self._get_nth_weekday_of_month(3, SUN, SEP))

    def _add_subdiv_barletta_holidays(self):
        self._add_holiday("San Ruggero", DEC, 30)

    def _add_subdiv_cesena_holidays(self):
        self._add_saint_johns_day("San Giovanni Battista")

    def _add_subdiv_forli_holidays(self):
        self._add_holiday("Madonna del Fuoco", FEB, 4)

    def _add_subdiv_pesaro_holidays(self):
        self._add_holiday("San Terenzio di Pesaro", SEP, 24)

    def _add_subdiv_trani_holidays(self):
        self._add_holiday("San Nicola Pellegrino", MAY, 3)

    def _add_subdiv_urbino_holidays(self):
        self._add_holiday("San Crescentino", JUN, 1)


class IT(Italy):
    pass


class ITA(Italy):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import MAY, JUN, AUG, SEP
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class ElSalvador(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    References:
    - https://www.transparencia.gob.sv/institutions/gd-usulutan/documents/192280/download
    - https://www.timeanddate.com/holidays/el-salvador
    - https://www.officeholidays.com/countries/el-salvador
    """

    country = "SV"
    subdivisions = (
        "AH",  # Ahuachapán
        "CA",  # Cabañas
        "CH",  # Chalatenango
        "CU",  # Cuscatlán
        "LI",  # La Libertad
        "MO",  # Morazán
        "PA",  # La Paz
        "SA",  # Santa Ana
        "SM",  # San Miguel
        "SO",  # Sonsonate
        "SS",  # San Salvador
        "SV",  # San Vicente
        "UN",  # La Unión
        "US",  # Usulután
    )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Maundy Thursday.
        self._add_holy_thursday("Maundy Thursday")

        # Good Friday.
        self._add_good_friday("Good Friday")

        # Holy Saturday.
        self._add_holy_saturday("Holy Saturday")

        # Labor Day.
        self._add_labor_day("Labor Day")

        if year >= 2016:
            # Legislative Decree #399 from Apr 14, 2016
            # Mothers' Day.
            self._add_holiday("Mothers' Day", MAY, 10)

        if year >= 2013:
            # Legislative Decree #208 from Jun 17, 2012
            # Fathers' Day.
            self._add_holiday("Fathers' Day", JUN, 17)

        # Feast of San Salvador.
        self._add_holiday("Feast of San Salvador", AUG, 6)

        # Independence Day.
        self._add_holiday("Independence Day", SEP, 15)

        # All Souls' Day.
        self._add_all_souls_day("All Souls' Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

    def _add_subdiv_ss_holidays(self):
        # San Salvador Day 1.
        self._add_holiday("San Salvador Day 1", AUG, 3)

        # San Salvador Day 2.
        self._add_holiday("San Salvador Day 2", AUG, 5)


class SV(ElSalvador):
    pass


class SLV(ElSalvador):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import MAR, APR, JUN, AUG, SEP, DEC, THU, SUN
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Switzerland(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "CH"
    default_language = "de"
    subdivisions = (
        "AG",  # Aargau
        "AR",  # Appenzell Ausserrhoden
        "AI",  # Appenzell Innerrhoden
        "BL",  # Basel-Landschaft
        "BS",  # Basel-Stadt
        "BE",  # Bern
        "FR",  # Fribourg
        "GE",  # Geneva
        "GL",  # Glarus
        "GR",  # Graubünden
        "JU",  # Jura
        "LU",  # Luzern
        "NE",  # Neuchâtel
        "NW",  # Nidwalden
        "OW",  # Obwalden
        "SG",  # St. Gallen
        "SH",  # Schaffhausen
        "SZ",  # Schwyz
        "SO",  # Solothurn
        "TG",  # Thurgau
        "TI",  # Ticino
        "UR",  # Uri
        "VD",  # Vaud
        "VS",  # Valais
        "ZG",  # Zug
        "ZH",  # Zürich
    )
    supported_languages = ("de", "en_US", "fr", "it", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Neujahrestag"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Ostern"))

        # Ascension Day.
        self._add_ascension_thursday(tr("Auffahrt"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Pfingsten"))

        if year >= 1291:
            # National Day.
            self._add_holiday(tr("Nationalfeiertag"), AUG, 1)

        # Christmas Day.
        self._add_christmas_day(tr("Weihnachten"))

    def _add_subdiv_holidays(self):
        if not self.subdiv:
            return None

        if self.subdiv != "VS":
            # Good Friday.
            self._add_good_friday(tr("Karfreitag"))

            # Easter Monday.
            self._add_easter_monday(tr("Ostermontag"))

            # Whit Monday.
            self._add_whit_monday(tr("Pfingstmontag"))

        if self.subdiv not in {"GE", "JU", "NE", "VD", "VS"}:
            # St. Stephen's Day.
            self._add_christmas_day_two(tr("Stephanstag"))

        super()._add_subdiv_holidays()

    def _add_subdiv_ag_holidays(self):
        # Berchtold's Day.
        self._add_new_years_day_two(tr("Berchtoldstag"))

    def _add_subdiv_ar_holidays(self):
        pass

    def _add_subdiv_ai_holidays(self):
        # Corpus Christi.
        self._add_corpus_christi_day(tr("Fronleichnam"))

        # Assumption of Mary.
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))

        # All Saints' Day.
        self._add_all_saints_day(tr("Allerheiligen"))

        # Immaculate Conception.
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_bl_holidays(self):
        # Labor Day.
        self._add_labor_day(tr("Tag der Arbeit"))

    def _add_subdiv_bs_holidays(self):
        self._add_labor_day(tr("Tag der Arbeit"))

    def _add_subdiv_be_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))

    def _add_subdiv_fr_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))

    def _add_subdiv_ge_holidays(self):
        # Thursday after the first Sunday of September
        self._add_holiday(
            # Genevan Fast.
            tr("Genfer Bettag"),
            self._get_nth_weekday_of_month(1, SUN, SEP) + td(days=+4),
        )

        # Restoration Day.
        self._add_holiday(tr("Wiederherstellung der Republik"), DEC, 31)

    def _add_subdiv_gl_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))

        # Näfelser Fahrt (first Thursday in April but not in Holy Week)
        if self._year >= 1835:
            dt = self._get_nth_weekday_of_month(1, THU, APR)
            if dt == self._easter_sunday + td(days=-3):
                dt += td(days=+7)
            # Battle of Naefels Victory Day.
            self._add_holiday(tr("Näfelser Fahrt"), dt)

        self._add_all_saints_day(tr("Allerheiligen"))

    def _add_subdiv_gr_holidays(self):
        pass

    def _add_subdiv_ju_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_labor_day(tr("Tag der Arbeit"))
        self._add_corpus_christi_day(tr("Fronleichnam"))

        # Independence Day.
        self._add_holiday(tr("Fest der Unabhängigkeit"), JUN, 23)
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))

    def _add_subdiv_lu_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_ne_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))

        # Republic Day.
        self._add_holiday(tr("Jahrestag der Ausrufung der Republik"), MAR, 1)
        self._add_labor_day(tr("Tag der Arbeit"))

        if self._is_sunday(self._christmas_day):
            self._add_christmas_day_two(tr("Stephanstag"))

    def _add_subdiv_nw_holidays(self):
        # St. Joseph's Day.
        self._add_saint_josephs_day(tr("Josefstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_ow_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))

        # St. Nicholas of Flüe.
        self._add_holiday(tr("Bruder Klaus"), SEP, 25)
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_sg_holidays(self):
        self._add_all_saints_day(tr("Allerheiligen"))

    def _add_subdiv_sh_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_labor_day(tr("Tag der Arbeit"))

    def _add_subdiv_sz_holidays(self):
        # Epiphany.
        self._add_epiphany_day(tr("Heilige Drei Könige"))
        self._add_saint_josephs_day(tr("Josefstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_so_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_labor_day(tr("Tag der Arbeit"))

    def _add_subdiv_tg_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_labor_day(tr("Tag der Arbeit"))

    def _add_subdiv_ti_holidays(self):
        self._add_epiphany_day(tr("Heilige Drei Könige"))
        self._add_saint_josephs_day(tr("Josefstag"))
        self._add_labor_day(tr("Tag der Arbeit"))
        self._add_corpus_christi_day(tr("Fronleichnam"))

        # Saints Peter and Paul.
        self._add_saints_peter_and_paul_day(tr("Peter und Paul"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_ur_holidays(self):
        self._add_epiphany_day(tr("Heilige Drei Könige"))
        self._add_saint_josephs_day(tr("Josefstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_vd_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))

        # Monday after the third Sunday of September
        self._add_holiday(
            # Prayer Monday.
            tr("Bettagsmontag"),
            self._get_nth_weekday_of_month(3, SUN, SEP) + td(days=+1),
        )

    def _add_subdiv_vs_holidays(self):
        self._add_saint_josephs_day(tr("Josefstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_zg_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_corpus_christi_day(tr("Fronleichnam"))
        self._add_assumption_of_mary_day(tr("Mariä Himmelfahrt"))
        self._add_all_saints_day(tr("Allerheiligen"))
        self._add_immaculate_conception_day(tr("Mariä Empfängnis"))

    def _add_subdiv_zh_holidays(self):
        self._add_new_years_day_two(tr("Berchtoldstag"))
        self._add_labor_day(tr("Tag der Arbeit"))


class CH(Switzerland):
    pass


class CHE(Switzerland):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import JAN, MAY, JUN, OCT, NOV, DEC, FRI, SAT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Finland(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Finland
    """

    country = "FI"
    default_language = "fi"
    supported_languages = ("en_US", "fi", "sv", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Uudenvuodenpäivä"))

        # Epiphany.
        name = tr("Loppiainen")
        if 1973 <= year <= 1990:
            self._add_holiday(name, self._get_nth_weekday_from(1, SAT, JAN, 6))
        else:
            self._add_epiphany_day(name)

        # Good Friday.
        self._add_good_friday(tr("Pitkäperjantai"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Pääsiäispäivä"))

        # Easter Monday.
        self._add_easter_monday(tr("2. pääsiäispäivä"))

        # May Day.
        self._add_holiday(tr("Vappu"), MAY, 1)

        # Ascension Day.
        name = tr("Helatorstai")
        if 1973 <= year <= 1990:
            self._add_holiday(name, self._easter_sunday + td(days=+34))
        else:
            self._add_ascension_thursday(name)

        # Whit Sunday.
        self._add_whit_sunday(tr("Helluntaipäivä"))

        dt = self._get_nth_weekday_from(1, FRI, JUN, 19) if year >= 1955 else date(year, JUN, 23)
        # Midsummer Eve.
        self._add_holiday(tr("Juhannusaatto"), dt)
        # Midsummer Day.
        self._add_holiday(tr("Juhannuspäivä"), dt + td(days=+1))

        dt = self._get_nth_weekday_from(1, SAT, OCT, 31) if year >= 1955 else date(year, NOV, 1)
        # All Saints' Day.
        self._add_holiday(tr("Pyhäinpäivä"), dt)

        # Independence Day.
        self._add_holiday(tr("Itsenäisyyspäivä"), DEC, 6)

        # Christmas Eve.
        self._add_christmas_eve(tr("Jouluaatto"))

        # Christmas Day.
        self._add_christmas_day(tr("Joulupäivä"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Tapaninpäivä"))


class FI(Finland):
    pass


class FIN(Finland):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import MAR, SEP, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays, IslamicHolidays


class Uzbekistan(HolidayBase, InternationalHolidays, IslamicHolidays):
    """
    https://www.officeholidays.com/countries/uzbekistan
    """

    country = "UZ"

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year
        self._add_new_years_day("New Year")

        # Women's Day
        self._add_womens_day("Women's Day")

        # Nauryz Holiday
        self._add_holiday("Nauryz", MAR, 21)

        # Ramadan Khait
        # Date of observance is announced yearly, This is an estimate.
        self._add_eid_al_fitr_day("Ramadan Khait")

        # Memorial Day
        self._add_world_war_two_victory_day("Memorial Day")

        # Kurban Khait
        # Date of observance is announced yearly, This is an estimate.
        self._add_eid_al_adha_day("Kurban Khait")

        # Independence Day
        self._add_holiday("Independence Day", SEP, 1)

        # Teacher's Day
        self._add_holiday("Teacher's Day", OCT, 1)

        # Constitution Day
        self._add_holiday("Constitution Day", DEC, 8)


class UZ(Uzbekistan):
    pass


class UZB(Uzbekistan):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars import _CustomIslamicCalendar
from holidays.calendars.gregorian import JAN, MAR, APR, MAY, JUN, JUL, AUG, SEP, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import InternationalHolidays, IslamicHolidays


class Azerbaijan(HolidayBase, InternationalHolidays, IslamicHolidays):
    # [1] https://en.wikipedia.org/wiki/Public_holidays_in_Azerbaijan
    # [2] https://az.wikipedia.org/wiki/Az%C9%99rbaycan%C4%B1n_d%C3%B6vl%C9%99t_bayramlar%C4%B1_v%C9%99_x%C3%BCsusi_g%C3%BCnl%C9%99ri  # noqa: E501
    # [3] https://www.sosial.gov.az/en/prod-calendar

    country = "AZ"

    def __init__(self, *args, **kwargs):
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self, calendar=AzerbaijanIslamicCalendar())
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        def _add_observed(dt: date, name: str = None):
            """
            Add observed holiday on next working day after specified date.
            """

            next_workday = dt + td(days=+1)
            while next_workday in hol_dates or self._is_weekend(next_workday):
                next_workday += td(days=+1)
            if name:
                self._add_holiday(f"{name} (Observed)", next_workday)
            else:
                for h_name in self.get_list(dt):
                    self._add_holiday(f"{h_name} (Observed)", next_workday)
            hol_dates.add(next_workday)

        if year <= 1989:
            return None

        super()._populate(year)
        observed_dates = set()
        non_observed_dates = set()

        # New Year
        name = "New Year's Day"
        observed_dates.add(self._add_new_years_day(name))
        if year >= 2006:
            observed_dates.add(self._add_new_years_day_two(name))

        # Black January (without extending)
        if year >= 2000:
            non_observed_dates.add(self._add_holiday("Black January", JAN, 20))

        # International Women's Day
        observed_dates.add(self._add_womens_day("International Women's Day"))

        # Novruz
        if year >= 2007:
            for d in range(20, 25):
                observed_dates.add(self._add_holiday("Novruz", MAR, d))

        # Victory Day
        observed_dates.add(self._add_world_war_two_victory_day("Victory Day over Fascism"))

        # Republic Day
        if year >= 1992:
            observed_dates.add(
                self._add_holiday("Independence Day" if year >= 2021 else "Republic Day", MAY, 28)
            )

        # National Salvation Day
        if year >= 1997:
            observed_dates.add(self._add_holiday("National Salvation Day", JUN, 15))

        # Memorial Day (without extending)
        if year >= 2021:
            non_observed_dates.add(self._add_holiday("Memorial Day", SEP, 27))

        # Azerbaijan Armed Forces Day
        if year >= 1992:
            name = "Azerbaijan Armed Forces Day"
            if year <= 1997:
                self._add_holiday(name, OCT, 9)
            else:
                observed_dates.add(self._add_holiday(name, JUN, 26))

        # Independence Day
        if year <= 2005:
            self._add_holiday("Independence Day", OCT, 18)

        # Victory Day
        if year >= 2021:
            observed_dates.add(self._add_holiday("Victory Day", NOV, 8))

        # Flag Day
        if year >= 2010:
            observed_dates.add(self._add_holiday("Flag Day", NOV, 9))

        # International Solidarity Day of Azerbaijanis
        if year >= 1993:
            solidarity_name = "International Solidarity Day of Azerbaijanis"
            self._add_new_years_eve(solidarity_name)

        if year >= 1993:
            name = "Ramazan Bayrami"
            observed_dates.update(self._add_eid_al_fitr_day(name))
            observed_dates.update(self._add_eid_al_fitr_day_two(name))

            name = "Gurban Bayrami"
            observed_dates.update(self._add_eid_al_adha_day(name))
            observed_dates.update(self._add_eid_al_adha_day_two(name))

        # Article 105 of the Labor Code of the Republic of Azerbaijan states:
        # 5. If interweekly rest days and holidays that are not considered
        # working days overlap, that rest day is immediately transferred to
        # the next working day.
        if self.observed and year >= 2006:
            hol_dates = observed_dates.union(non_observed_dates)

            dt = date(year - 1, DEC, 31)
            if self._is_weekend(dt):
                _add_observed(dt, solidarity_name)

            # observed holidays special cases
            special_dates_obs = {2007: (JAN, 3), 2072: (JAN, 5)}
            if year in special_dates_obs:
                hol_dates.add(
                    self._add_holiday(
                        "Gurban Bayrami* (*estimated) (Observed)", *special_dates_obs[year]
                    )
                )

            for hol_date in sorted(observed_dates):
                if self._is_weekend(hol_date):
                    _add_observed(hol_date)

                # 6. If the holidays of Qurban and Ramadan coincide with
                # another holiday that is not considered a working day,
                # the next working day is considered a rest day.
                elif len(self.get_list(hol_date)) > 1 and hol_date not in non_observed_dates:
                    for hol_name in self.get_list(hol_date):
                        if "Bayrami" in hol_name:
                            _add_observed(hol_date, hol_name)


class AZ(Azerbaijan):
    pass


class AZE(Azerbaijan):
    pass


class AzerbaijanIslamicCalendar(_CustomIslamicCalendar):
    EID_AL_ADHA_DATES = {
        2011: (NOV, 6),
        2012: (OCT, 25),
        2013: (OCT, 15),
        2014: (OCT, 4),
        2015: (SEP, 24),
        2016: (SEP, 12),
        2017: (SEP, 1),
        2018: (AUG, 22),
        2019: (AUG, 12),
        2020: (JUL, 31),
        2021: (JUL, 20),
        2022: (JUL, 9),
        2023: (JUN, 28),
    }

    EID_AL_FITR_DATES = {
        2011: (AUG, 30),
        2012: (AUG, 19),
        2013: (AUG, 8),
        2014: (JUL, 28),
        2015: (JUL, 17),
        2016: (JUL, 6),
        2017: (JUN, 26),
        2018: (JUN, 15),
        2019: (JUN, 5),
        2020: (MAY, 24),
        2021: (MAY, 13),
        2022: (MAY, 2),
        2023: (APR, 21),
    }



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import MAR, APR, AUG, OCT, NOV, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Hungary(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Hungary
    Codification dates:
      - https://hvg.hu/gazdasag/20170307_Megszavaztak_munkaszuneti_nap_lett_a_nagypentek  # noqa
      - https://www.tankonyvtar.hu/hu/tartalom/historia/92-10/ch01.html#id496839
    """

    country = "HU"
    default_language = "hu"
    supported_languages = ("en_US", "hu", "uk")

    def _add_observed(
        self, dt: date, since: int = 2010, before: bool = True, after: bool = True
    ) -> None:
        if not self.observed or dt.year < since:
            return None
        if self._is_tuesday(dt) and before:
            # Day off before
            self._add_holiday(self.tr("%s előtti pihenőnap") % self[dt], dt + td(days=-1))
        elif self._is_thursday(dt) and after:
            # Day off after
            self._add_holiday(self.tr("%s utáni pihenőnap") % self[dt], dt + td(days=+1))

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        name = self.tr("Újév")
        self._add_observed(self._add_new_years_day(name), before=False, since=2014)

        # The last day of the year is an observed day off if New Year's Day
        # falls on a Tuesday.
        if self.observed and self._is_monday(DEC, 31) and year >= 2014:
            self._add_holiday(self.tr("%s előtti pihenőnap") % name, DEC, 31)

        if 1945 <= year <= 1950 or year >= 1989:
            # National Day.
            self._add_observed(self._add_holiday(tr("Nemzeti ünnep"), MAR, 15))

        if year >= 2017:
            # Good Friday.
            self._add_good_friday(tr("Nagypéntek"))

        # Easter.
        self._add_easter_sunday(tr("Húsvét"))

        if year != 1955:
            # Easter Monday.
            self._add_easter_monday(tr("Húsvét Hétfő"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Pünkösd"))

        if year <= 1952 or year >= 1992:
            # Whit Monday.
            self._add_whit_monday(tr("Pünkösdhétfő"))

        if year >= 1946:
            # Labor Day.
            name = tr("A Munka ünnepe")
            self._add_observed(self._add_labor_day(name))
            if 1950 <= year <= 1953:
                self._add_labor_day_two(name)

        self._add_observed(
            self._add_holiday(
                # Bread Day.
                tr("A kenyér ünnepe") if 1950 <= year <= 1989 else
                # State Foundation Day.
                tr("Az államalapítás ünnepe"),
                AUG,
                20,
            )
        )

        if year >= 1991:
            # National Day.
            self._add_observed(self._add_holiday(tr("Nemzeti ünnep"), OCT, 23))

        if year >= 1999:
            # All Saints' Day.
            self._add_observed(self._add_all_saints_day(tr("Mindenszentek")))

        # Christmas Day.
        self._add_christmas_day(tr("Karácsony"))

        if year != 1955:
            self._add_observed(
                # Second Day of Christmas.
                self._add_christmas_day_two(tr("Karácsony másnapja")),
                since=2013,
                before=False,
            )

        # Soviet era.
        if 1950 <= year <= 1989:
            # Proclamation of Soviet Republic Day.
            self._add_holiday(tr("A Tanácsköztársaság kikiáltásának ünnepe"), MAR, 21)

            # Liberation Day.
            self._add_holiday(tr("A felszabadulás ünnepe"), APR, 4)

            if year not in {1956, 1989}:
                # Great October Socialist Revolution Day.
                self._add_holiday(tr("A nagy októberi szocialista forradalom ünnepe"), NOV, 7)


class HU(Hungary):
    pass


class HUN(Hungary):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import (
    JAN,
    FEB,
    MAR,
    APR,
    JUN,
    JUL,
    SEP,
    OCT,
    NOV,
    DEC,
    MON,
    TUE,
    WED,
)
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class NewZealand(HolidayBase, ChristianHolidays, InternationalHolidays):
    country = "NZ"
    special_holidays = {
        2022: (SEP, 26, "Queen Elizabeth II Memorial Day"),
    }
    subdivisions = (
        # https://en.wikipedia.org/wiki/ISO_3166-2:NZ
        "AUK",  # Auckland / Tāmaki-makaurau
        "BOP",  # Bay of Plenty / Toi Moana
        "CAN",  # Canterbury / Waitaha
        "CIT",  # Chatham Islands Territory / Wharekauri
        "GIS",  # Gisborne / Te Tairāwhiti
        "HKB",  # Hawke's Bay / Te Matau a Māui
        "MBH",  # Marlborough
        "MWT",  # Manawatū Whanganui
        "NSN",  # Nelson / Whakatū
        "NTL",  # Northland / Te Tai tokerau
        "OTA",  # Otago / Ō Tākou
        "STL",  # Southland / Te Taiao Tonga
        "TAS",  # Tasman / Te tai o Aorere
        "TKI",  # Taranaki
        "WGN",  # Greater Wellington / Te Pane Matua Taiao
        "WKO",  # Waikato
        "WTC",  # West Coast / Te Tai o Poutini
    )

    _deprecated_subdivisions = (
        "Auckland",
        "Canterbury",
        "Chatham Islands",
        "Hawke's Bay",
        "Marlborough",
        "Nelson",
        "New Plymouth",
        "Northland",
        "Otago",
        "South Canterbury",
        "STC",
        "Southland",
        "Taranaki",
        "Waitangi",
        "Wellington",
        "West Coast",
        "Westland",  # Correct name is West Coast
        "WTL",  # Correct code is WTC
    )

    def _get_nearest_monday(self, *args) -> date:
        dt = date(self._year, *args)
        return self._get_nth_weekday_from(
            1 if self._is_friday(dt) or self._is_weekend(dt) else -1, MON, dt
        )

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_weekend(dt):
            self._add_holiday(
                "%s (Observed)" % self[dt], dt + td(days=+2 if self._is_saturday(dt) else days)
            )

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # Bank Holidays Act 1873
        # The Employment of Females Act 1873
        # Factories Act 1894
        # Industrial Conciliation and Arbitration Act 1894
        # Labour Day Act 1899
        # Anzac Day Act 1920, 1949, 1956
        # New Zealand Day Act 1973
        # Waitangi Day Act 1960, 1976
        # Sovereign's Birthday Observance Act 1937, 1952
        # Holidays Act 1981, 2003

        if year <= 1893:
            return None

        super()._populate(year)

        # New Year's Day
        self._add_observed(self._add_new_years_day("New Year's Day"), days=+2)
        self._add_observed(self._add_new_years_day_two("Day after New Year's Day"), days=+2)

        # Waitangi Day
        if year >= 1974:
            name = "Waitangi Day" if year >= 1977 else "New Zealand Day"
            feb_6 = self._add_holiday(name, FEB, 6)
            if year >= 2014:
                self._add_observed(feb_6)

        # Anzac Day
        if year >= 1921:
            apr_25 = self._add_holiday("Anzac Day", APR, 25)
            if year >= 2014:
                self._add_observed(apr_25)

        # Easter
        self._add_good_friday("Good Friday")
        self._add_easter_monday("Easter Monday")

        # Sovereign's Birthday
        if year >= 1902:
            name = "Queen's Birthday" if 1952 <= year <= 2022 else "King's Birthday"
            if year == 1952:
                dt = date(year, JUN, 2)  # Elizabeth II
            elif year >= 1938:
                dt = self._get_nth_weekday_of_month(1, MON, JUN)  # EII & GVI
            elif year == 1937:
                dt = date(year, JUN, 9)  # George VI
            elif year == 1936:
                dt = date(year, JUN, 23)  # Edward VIII
            elif year >= 1912:
                dt = date(year, JUN, 3)  # George V
            else:
                # http://paperspast.natlib.govt.nz/cgi-bin/paperspast?a=d&d=NZH19091110.2.67
                dt = date(year, NOV, 9)  # Edward VII
            self._add_holiday(name, dt)

        # Matariki
        dates_obs = {
            2022: (JUN, 24),
            2023: (JUL, 14),
            2024: (JUN, 28),
            2025: (JUN, 20),
            2026: (JUL, 10),
            2027: (JUN, 25),
            2028: (JUL, 14),
            2029: (JUL, 6),
            2030: (JUN, 21),
            2031: (JUL, 11),
            2032: (JUL, 2),
            2033: (JUN, 24),
            2034: (JUL, 7),
            2035: (JUN, 29),
            2036: (JUL, 18),
            2037: (JUL, 10),
            2038: (JUN, 25),
            2039: (JUL, 15),
            2040: (JUL, 6),
            2041: (JUL, 19),
            2042: (JUL, 11),
            2043: (JUL, 3),
            2044: (JUN, 24),
            2045: (JUL, 7),
            2046: (JUN, 29),
            2047: (JUL, 19),
            2048: (JUL, 3),
            2049: (JUN, 25),
            2050: (JUL, 15),
            2051: (JUN, 30),
            2052: (JUN, 21),
        }
        if year in dates_obs:
            self._add_holiday("Matariki", *dates_obs[year])

        # Labour Day
        if year >= 1900:
            dt = (
                self._get_nth_weekday_of_month(4, MON, OCT)
                if year >= 1910
                else self._get_nth_weekday_of_month(2, WED, OCT)
            )
            self._add_holiday("Labour Day", dt)

        # Christmas Day
        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)

        # Boxing Day
        self._add_observed(self._add_christmas_day_two("Boxing Day"), days=+2)

        if self.subdiv == "Auckland":
            self._add_subdiv_auk_holidays()
        elif self.subdiv == "Canterbury":
            self._add_subdiv_can_holidays()
        elif self.subdiv == "Chatham Islands":
            self._add_subdiv_cit_holidays()
        elif self.subdiv == "Hawke's Bay":
            self._add_subdiv_hkb_holidays()
        elif self.subdiv == "Marlborough":
            self._add_subdiv_mbh_holidays()
        elif self.subdiv == "Nelson":
            self._add_subdiv_nsn_holidays()
        elif self.subdiv == "Northland":
            self._add_subdiv_ntl_holidays()
        elif self.subdiv == "Otago":
            self._add_subdiv_ota_holidays()
        elif self.subdiv in {"New Plymouth", "Taranaki"}:
            self._add_subdiv_tki_holidays()
        elif self.subdiv == "South Canterbury":
            self._add_subdiv_stc_holidays()
        elif self.subdiv == "Southland":
            self._add_subdiv_stl_holidays()
        elif self.subdiv == "Wellington":
            self._add_subdiv_wgn_holidays()
        elif self.subdiv in {"West Coast", "WTL", "Westland"}:
            self._add_subdiv_wtc_holidays()

    def _add_subdiv_auk_holidays(self):
        self._add_holiday("Auckland Anniversary Day", self._get_nearest_monday(JAN, 29))

    def _add_subdiv_can_holidays(self):
        self._add_holiday(
            "Canterbury Anniversary Day",
            self._get_nth_weekday_of_month(1, TUE, NOV) + td(days=+10),
        )

    def _add_subdiv_cit_holidays(self):
        self._add_holiday("Chatham Islands Anniversary Day", self._get_nearest_monday(NOV, 30))

    def _add_subdiv_hkb_holidays(self):
        self._add_holiday(
            "Hawke's Bay Anniversary Day",
            self._get_nth_weekday_of_month(4, MON, OCT) + td(days=-3),
        )

    def _add_subdiv_mbh_holidays(self):
        self._add_holiday(
            "Marlborough Anniversary Day",
            self._get_nth_weekday_of_month(4, MON, OCT) + td(days=+7),
        )

    def _add_subdiv_nsn_holidays(self):
        self._add_holiday("Nelson Anniversary Day", self._get_nearest_monday(FEB, 1))

    def _add_subdiv_ntl_holidays(self):
        if 1964 <= self._year <= 1973:
            name = "Waitangi Day"
            dt = (FEB, 6)
        else:
            name = "Auckland Anniversary Day"
            dt = (JAN, 29)
        self._add_holiday(name, self._get_nearest_monday(*dt))

    def _add_subdiv_ota_holidays(self):
        # there is no easily determined single day of local observance?!?!
        dt = self._get_nearest_monday(MAR, 23)
        if dt == self._easter_sunday + td(days=+1):  # Avoid Easter Monday
            dt += td(days=+1)
        self._add_holiday("Otago Anniversary Day", dt)

    def _add_subdiv_stc_holidays(self):
        self._add_holiday(
            "South Canterbury Anniversary Day", self._get_nth_weekday_of_month(4, MON, SEP)
        )

    def _add_subdiv_stl_holidays(self):
        dt = (
            self._easter_sunday + td(days=+2)
            if self._year >= 2012
            else self._get_nearest_monday(JAN, 17)
        )
        self._add_holiday("Southland Anniversary Day", dt)

    def _add_subdiv_tki_holidays(self):
        self._add_holiday("Taranaki Anniversary Day", self._get_nth_weekday_of_month(2, MON, MAR))

    def _add_subdiv_wgn_holidays(self):
        self._add_holiday("Wellington Anniversary Day", self._get_nearest_monday(JAN, 22))

    def _add_subdiv_wtc_holidays(self):
        dt = (
            date(self._year, DEC, 5)
            if self._year == 2005  # special case?!?!
            else self._get_nearest_monday(DEC, 1)
        )
        self._add_holiday("West Coast Anniversary Day", dt)


class NZ(NewZealand):
    pass


class NZL(NewZealand):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from gettext import gettext as tr

from holidays.calendars.gregorian import MAY, JUN
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Luxembourg(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_Luxembourg
    """

    country = "LU"
    default_language = "lb"
    supported_languages = ("de", "en_US", "fr", "lb", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Neijoerschdag"))

        # Easter Monday.
        self._add_easter_monday(tr("Ouschterméindeg"))

        # Labor Day.
        self._add_labor_day(tr("Dag vun der Aarbecht"))

        if year >= 2019:
            # Europe Day.
            self._add_holiday(tr("Europadag"), MAY, 9)

        # Ascension Day.
        self._add_ascension_thursday(tr("Christi Himmelfaart"))

        # Whit Monday.
        self._add_whit_monday(tr("Péngschtméindeg"))

        # National Day.
        self._add_holiday(tr("Nationalfeierdag"), JUN, 23)

        # Assumption Day.
        self._add_assumption_of_mary_day(tr("Léiffrawëschdag"))

        # All Saints' Day.
        self._add_all_saints_day(tr("Allerhellgen"))

        # Christmas Day.
        self._add_christmas_day(tr("Chrëschtdag"))

        # St. Stephen's Day.
        self._add_christmas_day_two(tr("Stiefesdag"))


class LU(Luxembourg):
    pass


class LUX(Luxembourg):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from gettext import gettext as tr

from holidays.calendars.gregorian import JUL, AUG, OCT, DEC
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Peru(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Peru holidays.

    References:
    - https://www.gob.pe/feriados
    - https://es.wikipedia.org/wiki/Anexo:Días_feriados_en_el_Perú
    """

    country = "PE"
    default_language = "es"
    supported_languages = ("en_US", "es", "uk")

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Año Nuevo"))

        # Holy Thursday.
        self._add_holy_thursday(tr("Jueves Santo"))

        # Good Friday.
        self._add_good_friday(tr("Viernes Santo"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Domingo de Resurrección"))

        # Labor Day.
        self._add_labor_day(tr("Día del Trabajo"))

        # Feast of Saints Peter and Paul.
        self._add_saints_peter_and_paul_day(tr("San Pedro y San Pablo"))

        # Independence Day.
        self._add_holiday(tr("Día de la Independencia"), JUL, 28)

        # Great Military Parade Day.
        self._add_holiday(tr("Día de la Gran Parada Militar"), JUL, 29)

        if year >= 2022:
            # Battle of Junín.
            self._add_holiday(tr("Batalla de Junín"), AUG, 6)

        # Santa Rosa de Lima.
        self._add_holiday(tr("Santa Rosa de Lima"), AUG, 30)

        # Battle of Angamos.
        self._add_holiday(tr("Combate de Angamos"), OCT, 8)

        # All Saints Day.
        self._add_all_saints_day(tr("Todos Los Santos"))

        # Immaculate Conception.
        self._add_immaculate_conception_day(tr("Inmaculada Concepción"))

        if year >= 2022:
            # Battle of Ayacucho.
            self._add_holiday(tr("Batalla de Ayacucho"), DEC, 9)

        # Christmas Day.
        self._add_christmas_day(tr("Navidad del Señor"))


class PE(Peru):
    pass


class PER(Peru):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td
from gettext import gettext as tr

from holidays.calendars.gregorian import GREGORIAN_CALENDAR, JAN, APR, JUN, JUL, AUG, OCT, NOV
from holidays.calendars.julian import JULIAN_CALENDAR
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Ukraine(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Current holidays list:
    https://zakon1.rada.gov.ua/laws/show/322-08/paran454#n454
    """

    country = "UA"
    default_language = "uk"
    supported_languages = ("ar", "en_US", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self, JULIAN_CALENDAR)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        # The current set of holidays came into force in 1991
        if year <= 1990:
            return None

        # There is no holidays in Ukraine during the period of martial law
        # https://zakon.rada.gov.ua/laws/show/2136-20#n26
        # law is in force from March 15, 2022
        if year >= 2023:
            return None

        super()._populate(year)
        observed_dates = set()

        # New Year's Day.
        observed_dates.add(self._add_new_years_day(tr("Новий рік")))

        observed_dates.add(
            self._add_christmas_day(
                # Christmas (Julian calendar).
                tr("Різдво Христове (за юліанським календарем)")
            )
        )

        # International Women's Day.
        observed_dates.add(self._add_womens_day(tr("Міжнародний жіночий день")))

        # There is no holidays from March 15, 2022
        # https://zakon.rada.gov.ua/laws/show/2136-20#n26
        if year <= 2021:
            # Easter Sunday (Pascha).
            observed_dates.add(self._add_easter_sunday(tr("Великдень (Пасха)")))

            # Holy Trinity Day.
            observed_dates.add(self._add_whit_sunday(tr("Трійця")))

            name = (
                # Labour Day.
                tr("День праці")
                if year >= 2018
                # International Workers' Solidarity Day.
                else tr("День міжнародної солідарності трудящих")
            )
            observed_dates.add(self._add_labor_day(name))
            if year <= 2017:
                observed_dates.add(self._add_labor_day_two(name))

            name = (
                # Day of Victory over Nazism in World War II (Victory Day).
                tr("День перемоги над нацизмом у Другій світовій війні (День перемоги)")
                if year >= 2016
                # Victory Day.
                else tr("День перемоги")
            )
            observed_dates.add(self._add_world_war_two_victory_day(name))

            if year >= 1997:
                observed_dates.add(
                    # Day of the Constitution of Ukraine.
                    self._add_holiday(tr("День Конституції України"), JUN, 28)
                )

            # Independence Day.
            name = tr("День незалежності України")
            if year >= 1992:
                observed_dates.add(self._add_holiday(name, AUG, 24))
            else:
                self._add_holiday(name, JUL, 16)

            if year >= 2015:
                name = (
                    # Day of defenders of Ukraine.
                    tr("День захисників і захисниць України")
                    if year >= 2021
                    # Defender of Ukraine Day.
                    else tr("День захисника України")
                )
                observed_dates.add(self._add_holiday(name, OCT, 14))

            if year <= 1999:
                # Anniversary of the Great October Socialist Revolution.
                name = tr("Річниця Великої Жовтневої соціалістичної революції")
                observed_dates.add(self._add_holiday(name, NOV, 7))
                observed_dates.add(self._add_holiday(name, NOV, 8))

            if year >= 2017:
                observed_dates.add(
                    self._add_christmas_day(
                        # Christmas (Gregorian calendar).
                        tr("Різдво Христове (за григоріанським календарем)"),
                        GREGORIAN_CALENDAR,
                    )
                )

        # 27.01.1995: holiday on weekend move to next workday
        # https://zakon.rada.gov.ua/laws/show/35/95-вр
        # 10.01.1998: cancelled
        # https://zakon.rada.gov.ua/laws/show/785/97-вр
        # 23.04.1999: holiday on weekend move to next workday
        # https://zakon.rada.gov.ua/laws/show/576-14
        if self.observed:
            for dt in sorted(observed_dates):
                if self._is_weekend(dt) and (
                    date(1995, JAN, 27) <= dt <= date(1998, JAN, 9) or dt >= date(1999, APR, 23)
                ):
                    obs_date = dt + td(days=+2 if self._is_saturday(dt) else +1)
                    while obs_date in self:
                        obs_date += td(days=+1)
                    hol_name = self.tr("%s (вихідний)") % self[dt]
                    self._add_holiday(hol_name, obs_date)


class UA(Ukraine):
    pass


class UKR(Ukraine):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from holidays.calendars.gregorian import OCT
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChineseCalendarHolidays, InternationalHolidays


class China(HolidayBase, ChineseCalendarHolidays, InternationalHolidays):
    """
    https://en.wikipedia.org/wiki/Public_holidays_in_China
    """

    country = "CN"

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        if year <= 1949:
            return None
        super()._populate(year)

        self._add_new_years_day("New Year's Day")

        name = "Labour Day"
        self._add_labor_day(name)
        if 2000 <= year <= 2007:
            self._add_labor_day_two(name)
            self._add_labor_day_three(name)

        name = "Chinese New Year (Spring Festival)"
        self._add_chinese_new_years_day(name)
        self._add_chinese_new_years_day_two(name)
        if 2008 <= year <= 2013:
            self._add_chinese_new_years_eve(name)
        else:
            self._add_chinese_new_years_day_three(name)

        name = "National Day"
        self._add_holiday(name, OCT, 1)
        self._add_holiday(name, OCT, 2)
        if year >= 2000:
            self._add_holiday(name, OCT, 3)

        if year >= 2008:
            self._add_qingming_festival("Tomb-Sweeping Day")
            self._add_dragon_boat_festival("Dragon Boat Festival")
            self._add_mid_autumn_festival("Mid-Autumn Festival")


class CN(China):
    pass


class CHN(China):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)


from holidays.calendars.gregorian import FEB, APR, JUN, AUG, NOV, DEC, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import (
    ChineseCalendarHolidays,
    ChristianHolidays,
    InternationalHolidays,
    IslamicHolidays,
)


class Philippines(
    HolidayBase, ChineseCalendarHolidays, ChristianHolidays, InternationalHolidays, IslamicHolidays
):
    """
    Philippines holidays.

    References:
      - https://en.wikipedia.org/wiki/Public_holidays_in_the_Philippines
    """

    country = "PH"

    def __init__(self, *args, **kwargs):
        ChineseCalendarHolidays.__init__(self)
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        IslamicHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day("New Year's Day")

        # Chinese New Year.
        self._add_chinese_new_years_day("Chinese New Year")

        # People Power Anniversary.
        self._add_holiday("EDSA Revolution Anniversary", FEB, 25)

        # Day of Valor.
        self._add_holiday("Day of Valor", APR, 9)

        # Maundy Thursday.
        self._add_holy_thursday("Maundy Thursday")
        # Good Friday.
        self._add_good_friday("Good Friday")
        # Black Saturday.
        self._add_holy_saturday("Black Saturday")

        # Labour Day.
        self._add_labor_day("Labour Day")

        # Eid al-Fitr.
        self._add_eid_al_fitr_day("Eid'l Fitr")

        # Independence Day.
        self._add_holiday("Independence Day", JUN, 12)

        # Eid al-Adha.
        self._add_eid_al_adha_day("Eid'l Adha")

        # Ninoy Aquino Day.
        self._add_holiday("Ninoy Aquino Day", AUG, 21)

        # National Heroes Day.
        self._add_holiday("National Heroes Day", self._get_nth_weekday_of_month(-1, MON, AUG))

        # All Saints' Day.
        self._add_all_saints_day("All Saints' Day")

        # Bonifacio Day.
        self._add_holiday("Bonifacio Day", NOV, 30)

        # Immaculate Conception Day.
        self._add_immaculate_conception_day("Immaculate Conception Day")

        # Christmas Day.
        self._add_christmas_day("Christmas Day")

        # Rizal Day.
        self._add_holiday("Rizal Day", DEC, 30)

        # New Year's Eve.
        self._add_new_years_eve("New Year's Eve")


class PH(Philippines):
    pass


class PHL(Philippines):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from datetime import timedelta as td

from holidays.calendars.gregorian import JUL, SEP, MON
from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Botswana(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    https://www.gov.bw/public-holidays
    https://publicholidays.africa/botswana/2021-dates/
    https://www.timeanddate.com/holidays/botswana/
    http://www.ilo.org/dyn/travail/docs/1766/Public%20Holidays%20Act.pdf
    """

    country = "BW"
    special_holidays = {2019: (JUL, 2, "Public Holiday")}

    def __init__(self, *args, **kwargs) -> None:
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _add_observed(self, dt: date, days: int = +1) -> None:
        if self.observed and self._is_sunday(dt) and self._year >= 1995:
            self._add_holiday("%s (Observed)" % self[dt], dt + td(days=days))

    def _populate(self, year):
        if year <= 1965:
            return None

        super()._populate(year)

        self._add_observed(self._add_new_years_day("New Year's Day"), days=+2)
        self._add_observed(self._add_new_years_day_two("New Year's Day Holiday"))

        # Easter and easter related calculations
        self._add_good_friday("Good Friday")
        self._add_holy_saturday("Holy Saturday")
        self._add_easter_monday("Easter Monday")
        self._add_ascension_thursday("Ascension Day")

        may_1 = self._add_labor_day("Labour Day")
        self._add_observed(may_1)
        if self.observed and year >= 2016 and self._is_saturday(may_1):
            self._add_labor_day_three("Labour Day Holiday")

        self._add_observed(self._add_holiday("Sir Seretse Khama Day", JUL, 1))

        july_3rd_monday = self._add_holiday(
            "President's Day", self._get_nth_weekday_of_month(3, MON, JUL)
        )
        self._add_holiday("President's Day Holiday", july_3rd_monday + td(days=+1))

        sep_30 = self._add_holiday("Botswana Day", SEP, 30)
        self._add_observed(sep_30, days=+2)
        self._add_observed(self._add_holiday("Botswana Day Holiday", sep_30 + td(days=+1)))

        self._add_observed(self._add_christmas_day("Christmas Day"), days=+2)
        dec_26 = self._add_christmas_day_two("Boxing Day")
        self._add_observed(dec_26)

        if self.observed and year >= 2016 and self._is_saturday(dec_26):
            self._add_holiday("Boxing Day Holiday", dec_26 + td(days=+2))


class BW(Botswana):
    pass


class BWA(Botswana):
    pass



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <maurizio.montel@gmail.com> (c) 2017-2022
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import date
from typing import Tuple, Optional

from holidays.calendars import _BuddhistLunisolar


class BuddhistCalendarHolidays:
    """
    Buddhist lunisolar calendar holidays.
    """

    def __init__(self, calendar=_BuddhistLunisolar(), show_estimated=False) -> None:
        self._buddhist_calendar = calendar
        self._show_estimated = show_estimated

    def _add_buddhist_calendar_holiday(
        self, name: str, hol_date: Tuple[date, bool]
    ) -> Optional[date]:
        """
        Add Buddhist calendar holiday.

        Adds customizable estimation label to holiday name if holiday date
        is an estimation.
        """
        estimated_label = getattr(self, "estimated_label", "%s* (*estimated)")
        dt, is_estimated = hol_date

        return self._add_holiday(
            self.tr(estimated_label) % self.tr(name)
            if is_estimated and self._show_estimated
            else name,
            dt,
        )

    def _add_vesak(self, name) -> Optional[date]:
        """
        Add Vesak (15th day of the 4th lunar month).

        Vesak for Thailand, Laos, Singapore and Indonesia.
        https://en.wikipedia.org/wiki/Vesak
        """
        return self._add_buddhist_calendar_holiday(
            name, self._buddhist_calendar.vesak_date(self._year)
        )

    def _add_vesak_may(self, name) -> Optional[date]:
        """
        Add Vesak (on the day of the first full moon in May
        in the Gregorian calendar).

        Vesak for Sri Lanka, Nepal, India, Bangladesh and Malaysia.
        https://en.wikipedia.org/wiki/Vesak
        """
        return self._add_buddhist_calendar_holiday(
            name, self._buddhist_calendar.vesak_may_date(self._year)
        )



#  python-holidays
#  ---------------
#  A fast, efficient Python library for generating country, province and state
#  specific sets of holidays on the fly. It aims to make determining whether a
#  specific date is a holiday as fast and flexible as possible.
#
#  Authors: dr-prodigy <dr.prodigy.github@gmail.com> (c) 2017-2023
#           ryanss <ryanssdev@icloud.com> (c) 2014-2017
#  Website: https://github.com/dr-prodigy/python-holidays
#  License: MIT (see LICENSE file)

from datetime import timedelta as td
from gettext import gettext as tr

from holidays.holiday_base import HolidayBase
from holidays.holiday_groups import ChristianHolidays, InternationalHolidays


class Denmark(HolidayBase, ChristianHolidays, InternationalHolidays):
    """
    Denmark holidays.

    References:
    - https://en.wikipedia.org/wiki/Public_holidays_in_Denmark
    - https://www.ft.dk/samling/20222/lovforslag/l13/index.htm
    """

    country = "DK"
    default_language = "da"
    supported_languages = ("da", "en_US", "uk")

    def __init__(self, *args, **kwargs):
        ChristianHolidays.__init__(self)
        InternationalHolidays.__init__(self)
        super().__init__(*args, **kwargs)

    def _populate(self, year):
        super()._populate(year)

        # New Year's Day.
        self._add_new_years_day(tr("Nytårsdag"))

        # Holy Thursday.
        self._add_holy_thursday(tr("Skærtorsdag"))

        # Good Friday.
        self._add_good_friday(tr("Langfredag"))

        # Easter Sunday.
        self._add_easter_sunday(tr("Påskedag"))

        # Easter Monday.
        self._add_easter_monday(tr("Anden påskedag"))

        # See https://www.ft.dk/samling/20222/lovforslag/l13/index.htm
        if year <= 2023:
            # Great Day of Prayers.
            self._add_holiday(tr("Store bededag"), self._easter_sunday + td(days=+26))

        # Ascension Day.
        self._add_ascension_thursday(tr("Kristi himmelfartsdag"))

        # Whit Sunday.
        self._add_whit_sunday(tr("Pinsedag"))

        # Whit Monday.
        self._add_whit_monday(tr("Anden pinsedag"))

        # Christmas Day.
        self._add_christmas_day(tr("Juledag"))

        # Second Day of Christmas.
        self._add_christmas_day_two(tr("Anden juledag"))


class DK(Denmark):
    pass


class DNK(Denmark):
    pass
